package de.opalpiraten.scenepainter.gui.render;
import java.util.Hashtable;
import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.image.*;
import com.threerings.cast.*;
import com.threerings.media.image.*;
import com.threerings.media.util.*;
import de.opalpiraten.scenepainter.sys.*;
import com.threerings.util.*;

/**
 * This class handles composite image elements loaded from nenya
 * media lib. Composite elements are rather complex. They consist
 * of several image resources, which can each have several coloring
 * objects bound to them.
 */
public class ComponentSceneElement extends SceneElement {

  /**
   * Speedup things: Cache a graphics context
   */
  private static final Graphics2D context=new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB).createGraphics();

  /**
   * Stores the resources required to build this element. This is
   * a Hashtable of Hashtables.
   */
  private Hashtable resources;
  
  /**
   * The action of the component
   */
  private String action;
  
  /**
   * The direction, the component is facing
   */
  private int orientation;
  
  /**
   * The index of the component image
   */
  private int index;
  
  /**
   * Number of frames available for this component
   */
  private int fcount=0;
  
  /**
   * Number of orientations supported by this component
   */
  private int ocount=0;
  
  
  public ComponentSceneElement() {}
  
  
  /**
   * Set the resources, this element consists of
   * @param hr A Hashtable of Hashtables. The hr Hashtable itself
   * is suppose to hold the names of the resourcenames (as they are
   * known to nenya), this ComponentSceneElement consists of as keys.
   * the values of those names are another Hashtable, containing a
   * mapping from colorclassnames to colornames. Hence this builds
   * a tree like structure. This structure was choosen, because it is
   * fairly easy to serialize via beans API and makes sure, that the
   * user cannot inadvently bind more than one color per class to
   * a ComponentResourceIdentifier (CRI).
   */
  public void setResources(Hashtable hr) {
    resources=hr;
    cache=null;
  }
  
  /**
   * Query the resources, this element consists of
   */
  public Hashtable getResources() {
    return resources;
  }
  
  /**
   * Set the action of the Composite
   */
  public void setAction(String a) {
    action=a;
    cache=null;
  }
  
  /**
   * Query the action of the Composite
   */
  public String getAction() {
    return action;
  } 
 
  /**
   * Set the Orientation, the component is facing
   */
  public void setOrientation(int o) {
    orientation =o;
    cache=null;
  }
  
  /**
   * Query Orientation
   */
  public int getOrientation() {
    return orientation;
  }
  
  /**
   * Set the index of the Frame to show
   */
  public void setIndex(int i) {
    index=i;
    cache=null;
  }
  
  /**
   * Query the index of the frame showing
   */
  public int getIndex() {
    return index;
  }
  
  public void scrollView(boolean forward) {
    int ticks=2;
    if (ocount==DirectionCodes.FINE_DIRECTION_COUNT) ticks=1;
    if (forward) orientation = DirectionUtil.rotateCW(orientation,ticks);
    else orientation= DirectionUtil.rotateCCW(orientation,ticks);
    cache=null;
  }
  
  
  // Implementation of abstract method
  public BufferedImage buildCache() {
    BufferedImage ret = null;
    
    // This is not the best way to do it, but nenya is a bit stubborn
    // when it comes to getting information. So reemplement a lot of stuff.
    try {
      if (action==null) throw new IllegalArgumentException("Action is null");
      if (resources==null) throw new IllegalArgumentException("resources is null");
      

      // First unpack the components
      CharacterComponent[] cc = new CharacterComponent[resources.size()];
      Enumeration e = resources.keys();
      int i=0;
      ComponentRepository cr = Core.nenyabrowser.getComponentRepository();
      while (e.hasMoreElements()) {
        cc[i]=cr.getComponent(Core.nenyabrowser.getComponentId((String)e.nextElement()));
        i++;
      }
      ActionFrames af = cc[0].getFrames(action,null);
      sort(cc,orientation,action);
      
      BufferedImage tmp=null;
      Graphics2D context=null;
      ColorPository cpos = project.getColorPository();
      
      TrimmedMultiFrameImage tmfi=null;
      
      // Colorize and paint every component
      for (i=0;i<cc.length;i++) {
        af = cc[i].getFrames(action,null);
        
        // Unpack the colors
        Hashtable colz = (Hashtable)resources.get(cc[i].componentClass.name+"/"+cc[i].name);
        Enumeration k = colz.keys();
        Vector collect = new Vector();
        while(k.hasMoreElements()) {
          String clazzname = (String) k.nextElement();
          String colorname = (String) colz.get(clazzname);
          collect.add(cpos.getColorization(clazzname,colorname));
        }
        Colorization[] collect_array = new Colorization[collect.size()];
        collect.copyInto(collect_array);
        af = af.cloneColorized(collect_array);
        
        tmfi= af.getFrames(orientation);
        
        
        if (tmp==null) {
          tmp=new BufferedImage(tmfi.getWidth(index),tmfi.getHeight(index),BufferedImage.TYPE_INT_ARGB);
          context = tmp.createGraphics();
        }
        tmfi.paintFrame(context,index,0,0);
      }
      // Everything needed is there -> assemble
      ocount=af.getOrientationCount();
      fcount= tmfi.getFrameCount();   
      ret= new BufferedImage(tmp.getWidth(),tmp.getHeight(),BufferedImage.TYPE_INT_ARGB);
      Graphics2D gr = ret.createGraphics();
      gr.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,alpha));
      gr.drawImage(tmp,0,0,null);
    }
    catch (Exception e) {
      //e.printStackTrace();
      // System.err.println(e.getMessage());
      ret = new BufferedImage(50,50,BufferedImage.TYPE_INT_ARGB);
      Graphics2D g=ret.createGraphics();
      g.setColor(Color.RED.brighter());
      g.fillOval(0,0,50,50);
      g.setColor(Color.white);
      g.fillRect(5,20,40,10);
      ocount=0;
      fcount=0;
    }
    return ret;
  }
  
  /**
   * Sorts components by renderpriority
   */
  private void sort(CharacterComponent[] cc, int dir, String act) {
    // This is bubblesort, but since there are only a few number of
    // components to sort, this is ok.
    boolean sorted=true;
    do {
      sorted=true;
      for (int i=1; i<cc.length;i++) {
        if(cc[i-1].componentClass.getRenderPriority(act,dir)>cc[i].componentClass.getRenderPriority(act,dir)) {
          final CharacterComponent turn=cc[i];
          cc[i]=cc[i-1]; 
          cc[i-1]=turn; 
          sorted=false;
        }
      }
    }
    while(!sorted);
  }
  
}
